"""
# Description

This module contains utility functions to handle multiple `qrotor.system` calculations.
These are commonly used as a list of `System` objects.


# Index

| | |
| --- | --- |
| `as_list()`          | Ensures that a list only contains System objects |
| `save_energies()`    | Save the energy eigenvalues for all systems to a CSV |
| `save_splittings()`  | Save the tunnel splitting energies for all systems to a CSV |
| `save_summary()`     | Save a summary of some relevant parameters for all systems to a CSV |
| `get_energies()`     | Get the eigenvalues from all systems |
| `get_gridsizes()`    | Get all gridsizes |
| `get_runtimes()`     | Get all runtimes |
| `get_ideal_E()`      | Calculate the ideal energy for a specified level |
| `sort_by_gridsize()` | Sort systems by gridsize |
| `reduce_size()`      | Discard data that takes too much space |
| `summary()`          | Print a summary of a System or list of Systems |
| `list_tags()`        | Get a list with all system tags |
| `filter_tags()`      | Filter the systems with or without specific tags |

---
"""


from .system import System
from aton import txt
import pandas as pd


def as_list(systems) -> None:
    """Ensures that `systems` is a list of System objects.

    If it is a System, returns a list with that System as the only element.
    If it is neither a list nor a System,
    or if the list does not contain only System objects,
    it raises an error.
    """
    if isinstance(systems, System):
        systems = [systems]
    if not isinstance(systems, list):
        raise TypeError(f"Must be a System object or a list of systems, found instead: {type(systems)}")
    for i in systems:
        if not isinstance(i, System):
            raise TypeError(f"All items in the list must be System objects, found instead: {type(i)}")
    return systems


def save_energies(
        systems:list,
        comment:str='',
        filepath:str='qrotor_eigenvalues.csv',
        ) -> pd.DataFrame:
    """Save the energy eigenvalues for all `systems` to a qrotor_eigenvalues.csv file.

    Returns a Pandas Dataset with `System.comment` columns and `System.eigenvalues` values.

    The output file can be changed with `filepath`,
    or set to null to avoid saving the dataset.
    A `comment` can be included at the top of the file.
    Note that `System.comment` must not include commas (`,`).
    """
    systems = as_list(systems)
    version = systems[0].version
    E = {}
    # Find max length of eigenvalues
    max_len = max((len(s.eigenvalues) if s.eigenvalues is not None else 0) for s in systems)
    for s in systems:
        if s.eigenvalues is not None:
            # Filter out None values and replace with NaN
            valid_eigenvalues = [float('nan') if e is None else e for e in s.eigenvalues]
            padded_eigenvalues = valid_eigenvalues + [float('nan')] * (max_len - len(s.eigenvalues))
        else:
            padded_eigenvalues = [float('nan')] * max_len
        E[s.comment] = padded_eigenvalues
    df = pd.DataFrame(E)
    if not filepath:
        return df
    # Else save to file
    df.to_csv(filepath, sep=',', index=False)
    # Include a comment at the top of the file
    file_comment = f'## {comment}\n' if comment else f''
    file_comment += f'# Energy eigenvalues\n'
    file_comment += f'# Calculated with QRotor {version}\n'
    file_comment += f'# https://pablogila.github.io/qrotor\n#'
    txt.edit.insert_at(filepath, file_comment, 0)
    print(f'Energy eigenvalues saved to {filepath}')
    return df


def save_splittings(
    systems:list,
    comment:str='',
    filepath:str='qrotor_splittings.csv',
    ) -> pd.DataFrame:
    """Save the tunnel splitting energies for all `systems` to a qrotor_splittings.csv file.

    Returns a Pandas Dataset with `System.comment` columns and `System.splittings` values.

    The output file can be changed with `filepath`,
    or set to null to avoid saving the dataset.
    A `comment` can be included at the top of the file.
    Note that `System.comment` must not include commas (`,`).
    Different splitting lengths across systems are allowed - missing values will be NaN.
    """
    systems = as_list(systems)
    version = systems[0].version
    tunnelling_E = {}
    # Find max length of splittings
    max_len = max(len(s.splittings) for s in systems)
    for s in systems:  # Pad shorter splittings with NaN
        padded_splittings = s.splittings + [float('nan')] * (max_len - len(s.splittings))
        tunnelling_E[s.comment] = padded_splittings
    df = pd.DataFrame(tunnelling_E)
    if not filepath:
        return df
    # Else save to file
    df.to_csv(filepath, sep=',', index=False)
    # Include a comment at the top of the file 
    file_comment = f'## {comment}\n' if comment else f''
    file_comment += f'# Tunnel splitting energies\n'
    file_comment += f'# Calculated with QRotor {version}\n'
    file_comment += f'# https://pablogila.github.io/qrotor\n#'
    txt.edit.insert_at(filepath, file_comment, 0)
    print(f'Tunnel splitting energies saved to {filepath}')
    return df


def save_summary(
    systems:list,
    comment:str='',
    filepath:str='qrotor_summary.csv',
    ) -> pd.DataFrame:
    """Save a summary for all `systems` to a qrotor_summary.csv file.

    Produces one row per System with the columns:
    `comment`, `ZPE`, `E_activation`, `potential_max`, `1st_splitting`,
    `1st_excitation`, `B`, `degeneracy`, `gridsize`.

    Set `filepath` to null to just return the DataFrame.
    """
    systems = as_list(systems)
    version = systems[0].version
    rows = []
    for s in systems:
        eigenvalues = getattr(s, 'eigenvalues', None)
        if eigenvalues is not None and len(eigenvalues) > 0:
            first_val = eigenvalues[0]
            zpe = float('nan') if first_val is None else first_val
        else:
            zpe = float('nan')
        splittings = getattr(s, 'splittings', None)
        if splittings is not None and len(splittings) > 0:
            first_splitting = float('nan') if splittings[0] is None else splittings[0]
        else:
            first_splitting = float('nan')
        excitations = getattr(s, 'excitations', None)
        if excitations is not None and len(excitations) > 0:
            first_excitation = float('nan') if excitations[0] is None else excitations[0]
        else:
            first_excitation = float('nan')
        system_comment = getattr(s, 'comment', None)
        E_activation = getattr(s, 'E_activation', None)
        B = getattr(s, 'B', None)
        tags = getattr(s, 'tags', None)
        deg = getattr(s, 'deg', None)
        gridsize = getattr(s, 'gridsize', None)
        potential_max = getattr(s, 'potential_max', None)
        # Each row contains the following:
        rows.append({
            'comment': system_comment,
            'ZPE': zpe,
            'E_activation': E_activation,
            'potential_max': potential_max,
            '1st_splitting': first_splitting,
            '1st_excitation': first_excitation,
            'B': B,
            'degeneracy': deg,
            'gridsize': gridsize,
            'tags': tags,
        })
    # Save to file or just return df
    df = pd.DataFrame(rows)
    if not filepath:
        return df
    df.to_csv(filepath, sep=',', index=False)
    # Include a comment at the top of the file
    file_comment = f'## {comment}\n' if comment else ''
    file_comment += '# Summary of systems\n'
    file_comment += f'# Calculated with QRotor {version}\n'
    file_comment += '# https://pablogila.github.io/qrotor\n#'
    txt.edit.insert_at(filepath, file_comment, 0)
    print(f'Summary saved to {filepath}')
    return df


def get_energies(systems:list) -> list:
    """Get a list with all lists of eigenvalues from all systems.

    If no eigenvalues are present for a particular system, appends None.
    """
    systems = as_list(systems)
    energies = []
    for i in systems:
        if all(i.eigenvalues):
            energies.append(i.eigenvalues)
        else:
            energies.append(None)
    return energies


def get_gridsizes(systems:list) -> list:
    """Get a list with all gridsize values.

    If no gridsize value is present for a particular system, appends None.
    """
    systems = as_list(systems)
    gridsizes = []
    for i in systems:
        if i.gridsize:
            gridsizes.append(i.gridsize)
        elif any(i.potential_values):
            gridsizes.append(len(i.potential_values))
        else:
            gridsizes.append(None)
    return gridsizes


def get_runtimes(systems:list) -> list:
    """Returns a list with all runtime values.
    
    If no runtime value is present for a particular system, appends None.
    """
    systems = as_list(systems)
    runtimes = []
    for i in systems:
        if i.runtime:
            runtimes.append(i.runtime)
        else:
            runtimes.append(None)
    return runtimes


def get_ideal_E(E_level:int) -> int:
    """Calculates the ideal energy for a specified `E_level`.

    To be used in convergence tests with `potential_name = 'zero'`.
    """
    real_E_level = None
    if E_level % 2 == 0:
        real_E_level = E_level / 2
    else:
        real_E_level = (E_level + 1) / 2
    ideal_E = int(real_E_level ** 2)
    return ideal_E


def sort_by_gridsize(systems:list) -> list:
    """Sorts a list of System objects by `System.gridsize`."""
    systems = as_list(systems)
    systems = sorted(systems, key=lambda sys: sys.gridsize)
    return systems


def reduce_size(systems:list) -> list:
    """Discard data that takes too much space.

    Removes eigenvectors, potential values and grids,
    for all System values inside the `systems` list.
    """
    systems = as_list(systems)
    for dataset in systems:
        dataset = dataset.reduce_size()
    return systems


def summary(
        systems,
        verbose:bool=False
        ) -> None:
    """Print a summary of a System or list of Systems.
    
    Print extra info with `verbose=True`
    """
    print('--------------------')
    systems = as_list(systems)
    
    for system in systems:
        dictionary = system.summary()
        if verbose:
            for key, value in dictionary.items():
                print(f'{key:<24}', value)
        else:
            eigenvalues = system.eigenvalues if any(system.eigenvalues) else []
            extra = ''
            if len(system.eigenvalues) > 6:
                eigenvalues = eigenvalues[:6]
                extra = '...'
            print('comment         ' + str(system.comment))
            print('ZPE             ' + str(system.eigenvalues[0]))
            print('E activation    ' + str(system.E_activation))
            print('V max           ' + str(system.potential_max))
            print('1st splitting   ' + str(system.splittings[0]))
            print('1st excitation  ' + str(system.excitations[0]))
            print('B               ' + str(system.B))
            print('eigenvalues     ' + str([float(round(e, 4)) for e in eigenvalues]) + extra)
            print('tags            ' + str(system.tags))
            print('version         ' + str(system.version))
        print('--------------------')
    return None


def list_tags(systems:list) -> list:
    """Returns a list with all system tags."""
    systems = as_list(systems)
    tags = []
    for i in systems:
        # i.tags is guaranteed to exist and be a string (may be empty)
        system_tags = i.tags.split()
        for tag in system_tags:
            if tag not in tags:
                tags.append(tag)
    return tags


def filter_tags(
        systems:list,
        include:str='',
        exclude:str='',
        strict:bool=False,
        ) -> list:
    """Returns a filtered list of systems with or without specific tags.

    You can `include` or `exclude` any number of tags, separated by blank spaces.
    By default, the filters are triggered if any tag is found, i.e. *tag1 OR tag2*.
    Set `strict=True` to require all tags to match, i.e. *tag1 AND tag2*.
    """
    systems = as_list(systems)
    included_tags = include.split()
    excluded_tags = exclude.split()
    filtered_systems = []
    for i in systems:
        tags_found = list_tags(i)
        if excluded_tags:
            if strict and all(tag in tags_found for tag in excluded_tags):
                continue
            elif not strict and any(tag in tags_found for tag in excluded_tags):
                continue
        if included_tags:
            if strict and not all(tag in tags_found for tag in included_tags):
                continue
            elif not strict and not any(tag in tags_found for tag in included_tags):
                continue
        filtered_systems.append(i)
    return filtered_systems




"""
# Description

This module provides straightforward functions to plot QRotor data.


# Index

| | |
| --- | --- |
| `potential()`        | Potential values as a function of the angle |
| `energies()`         | Calculated eigenvalues |
| `reduced_energies()` | Reduced energies E/B as a function of the reduced potential V/B |
| `wavefunction()`     | Selected wavefunctions or squared wavefunctions of a system |
| `splittings()`       | Tunnel splitting energies of a list of systems |
| `convergence()`      | Energy convergence of a list of systems calculated with different parameters |

---
"""


from .system import System
from . import systems
from . import constants
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy
import aton.alias as alias


def potential(
        data,
        title:str=None,
        marker='',
        linestyle='-',
        cm:bool=False,
        normalize:bool=False,
        ylim:tuple=None,
        ) -> None:
    """Plot the potential values of `data` (System object, or list of systems).

    Title can be customized with `title`.
    If empty, system[0].comment will be used as title if no more comments are present.

    `marker` and `linestyle` can be a Matplotlib string or list of strings.
    Optionally, the Viridis colormap can be used with `cm = True`.

    Set `normalize = True` to normalize by their respective `qrotor.system.System.potential_max`.
    This can be useful if you have performed subtractions or similar operations.
    In this case, you might also want to play with `ylim` to adjust the y-axis limits.
    """
    data_copy = deepcopy(data)
    system = systems.as_list(data_copy)
    title_str = title if title else (system[0].comment if (system[0].comment and (len(system) == 1 or not system[-1].comment)) else 'Rotational potential energy')
    # Marker as a list
    if isinstance(marker, list):
        if len(marker) < len(system):
            marker.extend([''] * (len(system) - len(marker)))
    else:
        marker = [marker] * len(system)
    # Linestyle as a list
    if isinstance(linestyle, list):
        if len(linestyle) < len(system):
            linestyle.extend(['-'] * (len(system) - len(linestyle)))
    else:
        linestyle = [linestyle] * len(system)

    plt.figure()
    plt.title(title_str)
    plt.xlabel('Angle / rad')
    plt.ylabel('Potential energy / meV')

    if normalize:
        plt.ylabel('Energy / V$_{3}$')
        for s in system:
            s.potential_values = s.potential_values / s.potential_max

    if ylim:
        plt.ylim(ylim)

    plt.xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])

    if cm:  # Plot using a colormap
        colors = plt.cm.viridis(np.linspace(0, 1, len(system)+1))  # +1 to avoid the lighter tones
        for i, s in enumerate(system):
            plt.plot(s.grid, s.potential_values, marker=marker[i], linestyle=linestyle[i], label=s.comment, color=colors[i])
    else:  # Regular plot
        for i, s in enumerate(system):
            plt.plot(s.grid, s.potential_values, marker=marker[i], linestyle=linestyle[i], label=s.comment)

    if all(s.comment for s in system) and len(system) != 1:
        plt.legend(fontsize='small')

    plt.show()


def energies(
        data,
        title:str=None,
        ) -> None:
    """Plot the eigenvalues of `data` (System or a list of System objects).

    You can use up to 1 tag per system to differentiate between molecular groups.
    """
    if isinstance(data, System):
        var = [data]
    else:  # Should be a list
        systems.as_list(data)
        var = data

    V_colors = ['C0', 'C1', 'C2', 'C3', 'C4']
    E_colors = ['lightblue', 'sandybrown', 'lightgrey', 'lightcoral', 'plum']
    E_linestyles = ['--', ':', '-.']
    edgecolors = E_colors

    V_linestyle = '-'
    title = title if title else (var[0].comment if var[0].comment else 'Energy eigenvalues')
    ylabel_text = f'Energy / meV'
    xlabel_text = 'Angle / radians'

    plt.figure(figsize=(10, 6))
    plt.xlabel(xlabel_text)
    plt.ylabel(ylabel_text)
    plt.title(title)
    plt.xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])

    unique_potentials = []
    unique_groups = []
    for i, system in enumerate(var):
        V_color = V_colors[i % len(V_colors)]
        E_color = E_colors[i % len(E_colors)]
        E_linestyle = E_linestyles[i % len(E_linestyles)]
        edgecolor = edgecolors[i % len(edgecolors)]

        # Plot potential energy if it is unique
        if not any(np.array_equal(system.potential_values, value) for value in unique_potentials):
            unique_potentials.append(system.potential_values)
            plt.plot(system.grid, system.potential_values, color=V_color, linestyle=V_linestyle)

        # Plot eigenvalues
        if any(system.eigenvalues):
            text_offset = 3 * len(unique_groups)
            if system.tags not in unique_groups:
                unique_groups.append(system.tags)
            for j, energy in enumerate(system.eigenvalues):
                plt.axhline(y=energy, color=E_color, linestyle=E_linestyle)
                # Textbox positions are a bit weird when plotting more than 2 systems, but whatever...
                plt.text(j%3*1.0 + text_offset, energy, f'$E_{{{j}}}$ = {round(energy,4):.04f}', va='top', bbox=dict(edgecolor=edgecolor, boxstyle='round,pad=0.2', facecolor='white', alpha=0.8))
            if len(systems.get_groups(var)) > 1:
                plt.plot([], [], color=E_color, label=f'{system.tags} Energies')  # Add to legend

    if len(systems.get_groups(var)) > 1:
        plt.subplots_adjust(right=0.85)
        plt.legend(bbox_to_anchor=(1.1, 0.5), loc='center', fontsize='small')

    plt.show()


def reduced_energies(
        data:list,
        title:str=None,
        values:list=[],
        legend:list=[],
        ) -> None:
    """Plots the reduced energy of the system E/B vs the reduced potential energy V/B.

    Takes a `data` list of System objects as input.
    An optional `title` can be specified.

    Optional maximum reduced potential `values` are plotted
    as vertical lines (floats or ints) or regions
    (lists inside the values list, from min to max).
    A `legend` of the same len as `values` can be included.
    These values are assumed to be divided by B by the user.
    """
    if values and (isinstance(values, float) or isinstance(values, int) or isinstance(values, np.float64)):
        values = [values]
    if values and len(values) <= len(legend):
        plot_legend = True
    else:
        plot_legend = False
        legend = [''] * len(values)
    systems.as_list(data)
    title = title if title else (data[0].comment if data[0].comment else 'Reduced energies')
    number_of_levels = data[0].searched_E
    x = []
    for system in data:
        potential_max_B = system.potential_max / system.B
        x.append(potential_max_B)
    colors = plt.cm.viridis(np.linspace(0, 1, number_of_levels+1))  # +1 to avoid the lighter tones
    for i in range(number_of_levels):
        y = []
        for system in data:
            eigenvalues_B_i = system.eigenvalues[i] / system.B
            y.append(eigenvalues_B_i)
        plt.plot(x, y, marker='', linestyle='-', color=colors[i])
    # Add vertical lines in the specified values
    line_colors = plt.cm.tab10(np.linspace(0, 1, len(values)))
    for i, value in enumerate(values):
        if isinstance(value, list):
            min_value = min(value)
            max_value = max(value)
            plt.axvspan(min_value, max_value, color=line_colors[i], alpha=0.2, linestyle='', label=legend[i])
        else:
            plt.axvline(x=value, color=line_colors[i], linestyle='--', label=legend[i], alpha=0.5)
    plt.xlabel('V$_{B}$ / B')
    plt.ylabel('E / B')
    plt.title(title)
    if plot_legend:
        plt.legend()
    plt.show()


def wavefunction(
        system:System,
        title:str=None,
        square:bool=True,
        levels=[0, 1, 2],
        overlap=False,
        yticks:bool=False,
        ) -> None:
    """Plot the wavefunction of a `system` for the specified `levels`.

    Wavefunctions are squared by default, showing the probabilities;
    To show the actual wavefunctions, set `square = False`.

    `levels` can be a list of indexes, or the number of levels to plot.

    Specific wavefunctions can be overlapped with `overlap` as a list with the target indexes.
    The `overlap` value can also be the max number of wavefunctions to add.
    All found wavefunctions can be added together with `overlap = True`;
    but note that this overlap is limited by the number of System.searched_E,
    that must be specified before solving the system.
    Setting `overlap` will ignore the `levels` argument.

    Set `yticks = True` to plot the wavefunction yticks.
    """
    data = deepcopy(system)
    eigenvectors = data.eigenvectors
    title = title if title else (data.comment if data.comment else 'System wavefunction')
    fig, ax1 = plt.subplots()
    plt.title(title)
    ax1.set_xlabel('Angle / radians')
    ax1.set_ylabel('Potential / meV')
    ax1.set_xticks([-2*np.pi, -3*np.pi/2, -np.pi, -np.pi/2, 0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi], [r'$-2\pi$', r'$-\frac{3\pi}{2}$', r'$-\pi$', r'$-\frac{\pi}{2}$', '0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])
    ax1.plot(data.grid, data.potential_values, color='blue', linestyle='-')
    ax2 = ax1.twinx()
    if not yticks:
        ax2.set_yticks([])
    ax2.set_ylabel('Squared wavefunction' if square else 'Wavefunction')
    # Set levels list
    if isinstance(levels, int) or isinstance(levels, float):
        levels = [x for x in range(int(levels))]
    if not isinstance(levels, list):
        raise ValueError('levels must be an int or a list of ints')
    # Set overlap if requested
    if overlap == True and isinstance(overlap, bool):
        eigenvectors = [np.sum(eigenvectors, axis=0)]
        levels = [0]
        show_legend = False
    elif overlap is not False and (isinstance(overlap, int) or isinstance(overlap, float)):
        max_int = int(overlap)
        eigenvectors = [np.sum(eigenvectors[:max_int], axis=0)]
        levels = [0]
        show_legend = False
    elif isinstance(overlap, list):
        eigenvectors = [np.sum([eigenvectors[i] for i in overlap], axis=0)]
        levels = [0]
        show_legend = False
    else:
        show_legend = True
    # Square values if so
    if square:
        eigenvectors = [vec**2 for vec in eigenvectors]
    # Plot the wavefunction
    for i in levels:
        ax2.plot(data.grid, eigenvectors[i], linestyle='--', label=f'{i}')
    if show_legend:
        fig.legend(loc='upper right', bbox_to_anchor=(0.9, 0.88), fontsize='small', title='Index')
    plt.show()


def splittings(
        data:list,
        title:str=None,
        units:str='ueV'
        ) -> None:
    """Plot the tunnel splitting energies of a `data` list of systems.

    The different `System.comment` are shown in the horizontal axis.
    An optional `title` can be specified.
    Default units shown are $\\mu$eV (`'ueV'`).
    Available units are: `'ueV'`, `'meV'`, `'Ry'`, or `'B'` (free rotor units).
    """
    title = title if title != None else 'Tunnel splitting energies'
    calcs = deepcopy(data)
    calcs = systems.as_list(calcs)

    fig, ax = plt.subplots()
    ax.set_ylabel("Energy / meV")

    y = [c.splittings[0] for c in calcs]
    x = [c.comment for c in calcs]
    # What units do we want?
    if units.lower() in alias.units['ueV']:
        y = [j * 1000 for j in y]  # Convert meV to micro eV
        ax.set_ylabel("Energy / $\\mu$eV")
    elif units.lower() in alias.units['Ry']:
        y = [j * constants.meV_to_Ry for j in y]
        ax.set_ylabel("Energy / Ry")
    elif units.upper() == 'B':
        y = [j / c.B for j, c in zip(y, calcs)]
        ax.set_ylabel("Energy / B")
    #else:  # It's okay let's use meV

    ax.bar(range(len(y)), y)
    for i, comment in enumerate(x):
        ax.text(x=i, y=0, s=comment+' ', rotation=45, verticalalignment='top', horizontalalignment='right')
    ax.set_xlabel("")
    ax.set_title(title)
    ax.set_xticks([])
    fig.tight_layout()
    plt.show()


def convergence(data:list) -> None:
    """Plot the energy convergence of a `data` list of Systems as a function of the gridsize."""
    systems.as_list(data)
    gridsizes = [system.gridsize for system in data]
    runtimes = [system.runtime for system in data]
    deviations = []  # List of lists, containing all eigenvalue deviations for every system
    searched_E = data[0].searched_E
    for system in data:
        deviation_list = []
        for i, eigenvalue in enumerate(system.eigenvalues):
            ideal_E = systems.get_ideal_E(i)
            deviation = abs(ideal_E - eigenvalue)
            deviation_list.append(deviation)
        deviation_list = deviation_list[1:]  # Remove ground state
        deviations.append(deviation_list)
    # Plotting
    fig, ax1 = plt.subplots()
    ax1.set_xlabel('Grid size')
    ax1.set_ylabel('Error / meV')
    ax1.set_xscale('log')
    ax1.set_yscale('log')
    ax2 = ax1.twinx()
    ax2.set_ylabel('Runtime / s')
    ax2.set_yscale('log')
    ax2.plot(gridsizes, runtimes, color='tab:grey', label='Runtime', linestyle='--')
    colors = plt.cm.viridis(np.linspace(0, 1, searched_E))  # Should be searched_E-1 but we want to avoid lighter colors
    for i in range(searched_E-1):
        if i % 2 == 0:  # Ignore even numbers, since those levels are degenerated.
            continue
        ax1.plot(gridsizes, [dev[i] for dev in deviations], label=f'$E_{{{int((i+1)/2)}}}$', color=colors[i])
    fig.legend(loc='upper right', bbox_to_anchor=(0.9, 0.88), fontsize='small')
    plt.title(data[0].comment if data[0].comment else 'Energy convergence vs grid size')
    plt.show()




"""
# Description

This module is used to solve any given quantum system.

Although the functions of this module can be used independently,
it is highly recommended to use the methods `System.solve()` or
`System.solve_potential()` instead to solve the whole quantum system
or just the potential values.
These user methods perform all calculations automatically,
see `qrotor.system.System.solve()` and
`qrotor.system.System.solve_potential()` respectively for more details.

This documentation page is left for reference and advanced users only.


# Index

| | |
| --- | --- |
| `energies()`              | Solve the quantum system, including eigenvalues and eigenvectors |
| `potential()`             | Solve the potential values of the system |
| `schrodinger()`           | Solve the Schrödiger equation for the system |
| `hamiltonian_matrix()`    | Calculate the hamiltonian matrix of the system |
| `laplacian_matrix()`      | Calculate the second derivative matrix for a given grid |
| `excitations()`           | Get excitation levels and tunnel splitting energies |
| `E_levels`                | Group a list of degenerated eigenvalues by energy levels |

---
"""


from .system import System
from .potential import solve as solve_potential
from .potential import interpolate
import time
import numpy as np
from scipy import sparse
import aton
from ._version import __version__


def energies(system:System, filename:str=None) -> System:
    """Solves the quantum `system`.

    This includes solving the potential, the eigenvalues and the eigenvectors.

    The resulting System object is saved with pickle to `filename` if specified.
    """
    system = potential(system)
    system = schrodinger(system)
    if filename:
        aton.file.save(system, filename)
    return system


def potential(system:System, gridsize:int=None) -> System:
    """Solves the potential values of the `system`.

    Creates a grid if not yet present.
    It also interpolates the potential if `system.gridsize` is larger than the current grid;
    optionally, an alternative `gridsize` can be specified.

    It then solves the potential according to the potential name.
    Then it applies extra operations, such as removing the potential offset
    if `system.correct_potential_offset = True`.
    """
    if gridsize:
        system.gridsize = gridsize
    if not any(system.grid):
        system.set_grid()
    if system.gridsize and any(system.grid):
        if system.gridsize > len(system.grid):
            system = interpolate(system)
    V = solve_potential(system)
    if system.correct_potential_offset is True:
        offset = min(V)
        V = V - offset
        system.potential_offset = offset
    system.potential_max = max(V)
    system.potential_min = min(V)
    system.potential_values = V
    return system


def schrodinger(system:System) -> System:
    """Solves the Schrödinger equation for a given `system`.
    
    Uses ARPACK in shift-inverse mode to solve the hamiltonian sparse matrix.
    """
    time_start = time.time()
    V = system.potential_values
    H = hamiltonian_matrix(system)
    print('Solving Schrodinger equation...')
    # Solve eigenvalues with ARPACK in shift-inverse mode, with a sparse matrix
    eigenvalues, eigenvectors = sparse.linalg.eigsh(H, system.searched_E, which='LM', sigma=0, maxiter=10000)
    if any(eigenvalues) is None:
        print('WARNING:  Not all eigenvalues were found.\n')
    else: print('Done.')
    system.version = __version__
    system.runtime = time.time() - time_start
    system.eigenvalues = eigenvalues
    system.E_activation = max(V) - min(eigenvalues)
    # Solve excitations and tunnel splittings, assuming triplet degeneracy
    system = excitations(system)
    # Do we really need to save eigenvectors?
    if system.save_eigenvectors == True:
        system.eigenvectors = np.transpose(eigenvectors)
    # Save potential max and min, in case these are not already saved
    system.potential_max = max(V)
    system.potential_min = min(V)
    return system


def hamiltonian_matrix(system:System):
    """Calculates the Hamiltonian sparse matrix for a given `system`."""
    print(f'Creating Hamiltonian sparse matrix of size {system.gridsize}...')
    V = system.potential_values.tolist()
    potential = sparse.diags(V, format='lil')
    B = system.B
    x = system.grid
    H = -B * laplacian_matrix(x) + potential
    return H


def laplacian_matrix(grid):
    """Calculates the Laplacian (second derivative) matrix for a given `grid`."""
    x = grid
    n = len(x)
    diagonals = [-2*np.ones(n), np.ones(n), np.ones(n)]
    laplacian_matrix = sparse.spdiags(diagonals, [0, -1, 1], m=n, n=n, format='lil')
    # Periodic boundary conditions
    laplacian_matrix[0, -1] = 1
    laplacian_matrix[-1, 0] = 1
    dx = x[1] - x[0]
    laplacian_matrix /= dx**2
    return laplacian_matrix


def excitations(system: System) -> System:
    """Calculate the excitation levels and the tunnel splitting energies of a system.

    Automatically detects degenerated energy levels by looking at significant jumps
    between consecutive eigenvalues. Within each level, finds two subgroups
    to calculate tunnel splittings. Stops when energies reach the maximum potential.

    Excitations are calculated as the energy difference between the mean energy of the
    ground state level and the mean energy of each excited level.

    Tunnel splittings are calculated as the difference between the mean values of
    the two subgroups within each degenerate level.
    """
    # Get eigenvalues, stop before any possible None value
    eigenvalues = system.eigenvalues
    if not isinstance(eigenvalues, (list, np.ndarray)) or len(eigenvalues) == 0:
        return system
    if None in eigenvalues:
        none_index = eigenvalues.tolist().index(None)
        eigenvalues = eigenvalues[:none_index]
    if len(eigenvalues) < 3:
        return system
    # Group degenerated eigenvalues into energy levels
    levels, degeneracy = E_levels(eigenvalues, system.potential_max)
    system.E_levels = levels
    system.deg = degeneracy
    # Calculate excitations and splittings
    ground_energy = np.mean(levels[0])  # Mean of ground state level
    excitations = []
    tunnel_splittings = []
    for level in levels:
        level_mean = np.mean(level)
        excitations.append(level_mean - ground_energy)
        # Get the tunnel splitting within the level
        if len(level) > 1:
            # Find the largest gap within the level to split into two subgroups
            internal_gaps = np.diff(level)
            split_idx = np.argmax(internal_gaps) + 1
            # Split into two subgroups
            subgroup1 = level[:split_idx]
            subgroup2 = level[split_idx:]
            # Medians of subgroups
            median1 = np.median(subgroup1)
            median2 = np.median(subgroup2)
            # Tunnel splitting is the difference between medians
            tunnel_splittings.append(abs(median2 - median1))
        else:
            tunnel_splittings.append(0)
    system.excitations = excitations[1:]  # Exclude ground state
    system.splittings = tunnel_splittings
    return system


def E_levels(eigenvalues, vmax:float=None) -> list:
    """Group a list of degenerated eigenvalues by energy levels.

    Automatically detects degenerated energy levels by
    looking at significant jumps between consecutive eigenvalues.

    An optional `vmax` can be specified,
    to avoid including too many eigenvalues
    above a certain potential maximum.
    Only two more eigenvalues are considered after `vmax`,
    to properly detect energy levels around the maximum.

    Example:
    ```python
    levels, deg = qr.solve.E_levels(array([1.1, 1.2, 1.3, 5.4, 5.5, 5.6]))
    levels  # [array([1.1, 1.2, 1.3]), array([5.4, 5.5, 5.6])]
    deg  # 3
    ```
    """
    if vmax:  # Include all eigenvalues below Vmax plus 3 more eigenvalues
        # Check if any values are above vmax
        eigenvalues_above_vmax = eigenvalues > vmax
        if np.any(eigenvalues_above_vmax):
            index_first_above_vmax = np.where(eigenvalues_above_vmax)[0][0]
            eigenvalues = eigenvalues[:(index_first_above_vmax + 2)]
    # Group degenerated eigenvalues into energy levels
    for scale in np.arange(2, 4, 0.25):  # First search going to bigger scales
        levels, degeneracy = _get_E_levels_by_gap(eigenvalues, scale)
        if (degeneracy > 1) and (degeneracy % 1 == 0):
            break
        else:
            levels, degeneracy = None, None
    if not degeneracy:  # If it didn't work, search with tighter values
        for scale in np.arange(0.75, 2, 0.25):
            levels, degeneracy = _get_E_levels_by_gap(eigenvalues, scale)
            if (degeneracy > 1) and (degeneracy % 1 == 0):
                break
    if not (degeneracy > 1) and not (degeneracy % 1) == 0:
        return levels, degeneracy  # I give up
    # Correct the last two levels
    if len(levels) >= 2 and len(levels[-2]) != degeneracy:
        levels[-2] = np.concatenate((levels[-2], levels[-1]))
        levels.pop(-1)
    # Split last level into groups of size = degeneracy
    last_level = levels[-1]
    additional_levels = len(last_level) // degeneracy
    if additional_levels > 0:
        # Replace last level with list of complete degeneracy groups
        complete_groups = [last_level[i:i+degeneracy] for i in range(0, additional_levels*degeneracy, degeneracy)]
        levels.pop(-1)  # Remove original last level
        levels.extend(complete_groups)  # Add all complete groups
    else:
        levels.pop(-1)  # Remove incomplete last level
    return levels, degeneracy


def _get_E_levels_by_gap(eigenvalues, scale:float=2) -> tuple:
    """Split a list of eigenvalues into energy levels by looking at gaps.
    
    If the gap is bigger than the average gap times `scale`, it is considered a new level.

    Returns a tuple with the estimated levels and the average degeneracy.
    The last two levels are not taken into account to estimate the degeneracy.
    """
    # Find gaps between consecutive eigenvalues
    gaps = np.diff(eigenvalues)
    # Use mean gap times scale as threshold to distinguish energy levels
    med_gap = np.mean(gaps)
    level_breaks = np.where(gaps > scale * med_gap)[0] + 1
    levels = np.split(eigenvalues, level_breaks)
    # Calculate average degeneracy excluding last two levels if possible
    if len(levels) > 2:
        avg_degeneracy = float(np.mean([len(level) for level in levels[:-2]]))
    else:
        avg_degeneracy = float(len(levels[0]))
    if avg_degeneracy % 1 == 0:
        avg_degeneracy = int(avg_degeneracy)
    return levels, avg_degeneracy




import qrotor as qr


def test_solve_zero():
    system = qr.System()
    system.gridsize = 50000
    system.potential_name = 'zero'
    system.B = 1
    system.solve()
    assert round(system.eigenvalues[0], 2) == 0.0
    assert round(system.eigenvalues[1], 2) == 1.0
    assert round(system.eigenvalues[2], 2) == 1.0
    assert round(system.eigenvalues[3], 2) == 4.0
    assert round(system.eigenvalues[4], 2) == 4.0
    assert round(system.eigenvalues[5], 2) == 9.0
    assert round(system.eigenvalues[6], 2) == 9.0
    assert round(system.eigenvalues[7], 2) == 16.0
    assert round(system.eigenvalues[8], 2) == 16.0


def test_solve_potential():
    system = qr.System()
    system.gridsize = 500
    system.potential_name = 'sin'
    system.potential_constants = [0, 1, 3, 0]
    system.solve_potential()
    assert round(system.potential_max, 2) == 1.0




"""
# Description

This submodule contains tools to rotate molecular structures.
Works with Quantum ESPRESSO input files.


# Index

| | |
| --- | --- |
| `rotate_qe()`     | Rotate specific atoms from a Quantum ESPRESSO input file |
| `rotate_coords()` | Rotate a specific list of coordinates |

---
"""


import numpy as np
import os
import shutil
from scipy.spatial.transform import Rotation
from .constants import *
import aton.api as api
import aton.txt.extract as extract
import aton.txt.edit as edit


def rotate_qe(
        filepath:str,
        positions:list,
        angle:float,
        repeat:bool=False,
        precision:int=3,
        use_centroid:bool=True,
        show_axis:bool=False,
    ) -> list:
    """Rotates atoms from a Quantum ESPRESSO pw.x input file.

    Takes a `filepath` with a molecular structure, and three or more atomic `positions` (list).
    These input positions can be approximate, and are used to identify the target atoms.
    The decimal precision in the search for these positions is controlled by `precision`.

    It rotates the atoms by a specific `angle` in degrees.
    Additionally, if `repeat = True` it repeats the same rotation over the whole circunference.
    Finally, it writes the rotated structure(s) to a new structural file(s).
    Returns a list with the output filename(s).

    By default, the rotation axis is defined by the perpendicular vector
    passing through the geometrical center of the first three points.
    To override this and instead use the vector between the first two atoms
    as the rotation axis, set `use_centroid = False`.

    **WARNING: The `positions` list is order-sensitive**.
    If you rotate more than one chemical group in a structure,
    be sure to follow the same direction for each group (e.g. all clockwise)
    to ensure that all axes of rotation point in the same direction.

    To debug, `show_axis = True` adds two additional helium atoms as the rotation vector.

    The resulting rotational potential can be compiled to a CSV file with `qrotor.potential.from_qe()`.
    """
    print('Rotating Quantum ESPRESSO input structure with QRotor...')
    if len(positions) < 3:
        raise ValueError("At least three positions are required to define the rotation axis.")
    lines = []
    full_positions = []
    for position in positions:
        line = api.pwx.get_atom(filepath, position, precision)
        lines.append(line)
        pos = extract.coords(line)
        if len(pos) > 3:  # Keep only the first three coordinates
            pos = pos[:3]
        # Convert to cartesian
        pos_cartesian = api.pwx.to_cartesian(filepath, pos)
        full_positions.append(pos_cartesian)
        print(f'Found atom: "{line}"')
    # Set the angles to rotate
    if not repeat:
        angles = [angle]
    else:
        angles = range(0, 360, angle)
    # Rotate and save the structure
    outputs = []
    path = os.path.dirname(filepath)
    basename = os.path.basename(filepath)
    name, ext = os.path.splitext(basename)
    print('Rotating the structure...')
    for angle in angles:
        output_name = name + f'_{angle}' + ext
        output = os.path.join(path, output_name)
        rotated_positions_cartesian = rotate_coords(full_positions, angle, use_centroid, show_axis)
        rotated_positions = []
        for coord in rotated_positions_cartesian:
            pos = api.pwx.from_cartesian(filepath, coord)
            rotated_positions.append(pos)
        _save_qe(filepath, output, lines, rotated_positions)
        outputs.append(output)
        print(output)
    return outputs


def rotate_coords(
        positions:list,
        angle:float,
        use_centroid:bool=True,
        show_axis:bool=False,
    ) -> list:
    """Rotates geometrical coordinates.

    Takes a list of atomic `positions` in cartesian coordinates, as
    `[[x1,y1,z1], [x2,y2,z2], [x3,y3,z3], [etc]`.
    Then rotates said coordinates by a given `angle` in degrees.
    Returns a list with the updated positions.

    By default, the rotation vector is defined by the perpendicular
    passing through the geometrical center of the first three points.
    To override this and use the vector between the first two atoms
    as the rotation axis, set `use_centroid = False`.

    **WARNING: The `positions` list is order-sensitive**.
    If you rotate more than one chemical group in a structure,
    be sure to follow the same direction for each group (e.g. all clockwise)
    to ensure that all rotation vectors point in the same direction.

    If `show_axis = True` it returns two additional coordinates at the end of the list,
    with the centroid and the rotation vector. Only works with `use_centroid = True`.

    The rotation uses Rodrigues' rotation formula,
    powered by [`scipy.spatial.transform.Rotation.from_rotvec`](https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.from_rotvec.html#scipy.spatial.transform.Rotation.from_rotvec).
    """
    if len(positions) < 3:
        raise ValueError("At least three atoms must be rotated.")
    if not isinstance(positions[0], list):
        raise ValueError(f"Atomic positions must have the form: [[x1,y1,z1], [x2,y2,z2], [x3,y3,z3], etc]. Yours were:\n{positions}")
    positions = np.array(positions)
    #print(f'POSITIONS: {positions}')  # DEBUG
    # Define the geometrical center
    center_atoms = positions[:2]
    if use_centroid:
        center_atoms = positions[:3]
    center = np.mean(center_atoms, axis=0)
    # Ensure the axis passes through the geometrical center
    centered_positions = positions - center
    # Define the perpendicular axis (normal to the plane formed by the first three points)
    v1 = centered_positions[0] - centered_positions[1]
    v2 = centered_positions[0] - centered_positions[2]
    axis = v1  # Axis defined by the first two points
    if use_centroid:  # Axis defined by the cross product of the first three points
        axis = np.cross(v2, v1)
    axis_length = np.linalg.norm(axis)
    axis = axis / axis_length
    # Create the rotation object using scipy
    rotation = Rotation.from_rotvec(np.radians(angle) * axis)
    # Rotate all coordinates around the geometrical center
    rotated_centered_positions = rotation.apply(centered_positions)
    rotated_positions = (rotated_centered_positions + center).tolist()
    #print(f'ROTATED_POSITIONS: {rotated_positions}')  # DEBUG
    if show_axis and use_centroid:
        rotated_positions.append(center.tolist())
        rotated_positions.append((center + axis).tolist())
    return rotated_positions


def _save_qe(
        filename,
        output:str,
        lines:list,
        positions:list
    ) -> str:
    """Copies `filename` to `output`, updating the old `lines` with the new `positions`.
    
    The angle will be appended at the end of the input prefix to avoid overlapping calculations.
    """
    shutil.copy(filename, output)
    for i, line in enumerate(lines):
        strings = line.split()
        atom = strings[0]
        new_line = f"  {atom}   {positions[i][0]:.15f}   {positions[i][1]:.15f}   {positions[i][2]:.15f}"
        #print(f'OLD LINE: {line}')  # DEBUG
        #print(f'NEW_LINE: {new_line}')  # DEBUG
        edit.replace_line(output, line, new_line, raise_errors=True)
    if len(lines) + 2 == len(positions):  # In case show_axis=True
        additional_positions = positions[-2:]
        for pos in additional_positions:
            pos.insert(0, 'He')
            api.pwx.add_atom(output, pos)
    elif len(lines) != len(positions):
        raise ValueError(f"What?!  len(lines)={len(lines)} and len(positions)={len(positions)}")
    # Add angle to calculation prefix
    output_name = os.path.basename(output)
    splits = output_name.split('_')
    angle_str = splits[-1].replace('.in', '')
    prefix = ''
    content = api.pwx.read_in(output)
    if 'prefix' in content.keys():
        prefix = content['prefix']
        prefix = prefix.strip("'")
    prefix = "'" + prefix + angle_str + "'"
    api.pwx.set_value(output, 'prefix', prefix)
    return output




"""
# Description

This module contains functions to calculate the actual `potential_values` of the system.


# Index

User functions:

| | |
| --- | --- |
| `save()`        | Save the potential from a System to a data file |
| `load()`        | Load a System with a custom potential from a potential data file |
| `from_qe()`     | Creates a potential data file from Quantum ESPRESSO outputs |
| `merge()`       | Add and subtract potentials from systems |
| `scale()`       | Scale potential values by a given factor |

To solve the system, optionally interpolating to a new gridsize, use the `System.solve(gridsize)` method.  
However, if you just want to quickly solve or interpolate the potential, check the `System.solve_potential(gridsize)` method.
This will run several checks before applying the following functions automatically:

| | |
| --- | --- |
| `interpolate()` | Interpolates the current `System.potential_values` to a new `System.gridsize` |
| `solve()`       | Solve the potential values based on the potential name |

A synthetic potential can be created by specifying its name in `System.potential_name`,
along with the corresponding `System.potential_constants` if required.
Available potentials are:

| | |
| --- | --- |
| `zero()`        | Zero potential |
| `sine()`        | Sine potential |
| `cosine()`      | Cosine potential |
| `titov2023()`   | Potential of the hindered methyl rotor, as in titov2023. |

---
"""


from .system import System
from . import constants
from . import systems
import numpy as np
import os
from copy import deepcopy
from scipy.interpolate import CubicSpline
import aton.alias as alias
import aton.file as file
import aton.api.pwx as pwx
from ._version import __version__


def save(
        system:System,
        comment:str='',
        filepath:str='potential.csv',
        angle:str='deg',
        energy:str='meV',
        ) -> None:
    """Save the rotational potential from a `system` to a CSV file.

    The output `filepath` contains angle and energy columns,
    in degrees and meVs by default.
    The units can be changed with `angle` and `energy`,
    but only change these defaults if you know what you are doing.
    An optional `comment` can be included in the header of the file.
    """
    print('Saving potential data file...')
    # Check if a previous potential.dat file exists, and ask to overwrite it
    previous_potential_file = file.get(filepath, return_anyway=True)
    if previous_potential_file:
        print(f"WARNING: Previous '{filepath}' file will be overwritten, proceed anyway?")
        answer = input("(y/n): ")
        if not answer.lower() in alias.boolean[True]:
            print("Aborted.")
            return None
    # Set header
    potential_data = f'## {comment}\n' if comment else f'## {system.comment}\n' if system.comment else ''
    potential_data += '# Rotational potential dataset\n'
    potential_data += f'# Saved with QRotor {__version__}\n'
    potential_data += '# https://pablogila.github.io/qrotor\n'
    potential_data += '#\n'
    # Check that grid and potential values are the same size
    if len(system.grid) != len(system.potential_values):
        raise ValueError('len(system.grid) != len(system.potential_values)')
    grid = system.grid
    potential_values = system.potential_values
    # Convert angle units
    if angle.lower() in alias.units['rad']:
        potential_data += '# Angle/rad,    '
    else:
        grid = np.degrees(grid)
        potential_data += '# Angle/deg,    '
        if not angle.lower() in alias.units['deg']:
            print(f"WARNING: Unrecognised '{angle}' angle units, using degrees instead")
    # Convert energy units
    if energy.lower() in alias.units['meV']:
        potential_data += 'Potential/meV\n'
    elif energy.lower() in alias.units['eV']:
        potential_values = potential_values * 1e-3
        potential_data += 'Potential/eV\n'
    elif energy.lower() in alias.units['Ry']:
        potential_values = potential_values * constants.meV_to_Ry
        potential_data += 'Potential/Ry\n'
    else:
        print(f"WARNING:  Unrecognised '{energy}' energy units, using meV instead")
        potential_data += 'Potential/meV\n'
    potential_data += '#\n'
    # Save all values
    for angle_value, energy_value in zip(grid, potential_values):
        potential_data += f'{angle_value},    {energy_value}\n'
    with open(filepath, 'w') as f:
        f.write(potential_data)
    print(f'Saved to {filepath}')
    # Warn the user if not in default units
    if angle.lower() not in alias.units['deg']:
        print(f"WARNING: You saved the potential in '{angle}' angle units! Remember that QRotor works in degrees!")
    if energy.lower() not in alias.units['meV']:
        print(f"WARNING: You saved the potential in '{energy}' energy units! Remember that QRotor works in meVs!")


def load(
        filepath:str='potential.csv',
        comment:str=None,
        tags:str='',
        system:System=None,
        angle:str='deg',
        energy:str='meV',
        ) -> System:
    """Read a rotational potential energy datafile.

    The input file in `filepath` should contain two columns with angle and potential energy values.
    Degrees and meV are assumed as default units unless stated in `angle` and `energy`.
    Units will be converted automatically to radians and meV.

    An optional `comment` can be included in the output System.
    Set to the parent folder name by default.

    A previous System object can be provided through `system` to update its potential values.
    """
    file_path = file.get(filepath)
    system = System() if system is None else system
    with open(file_path, 'r') as f:
        lines = f.readlines()
    # Read the comment
    loaded_comment = ''
    if lines[0].startswith('## '):
        loaded_comment = lines[0][3:].strip()
    # Read data
    positions = []
    potentials = []
    for line in lines:
        if line.startswith('#'):
            continue
        position, potential = line.split()
        positions.append(float(position.strip().strip(',').strip()))
        potentials.append(float(potential.strip()))
    # Save angles to numpy arrays
    if angle.lower() in alias.units['deg']:
        positions = np.radians(positions)
    elif angle.lower() in alias.units['rad']:
        positions = np.array(positions)
    else:
        raise ValueError(f"Angle unit '{angle}' not recognized.")
    # Save energies to numpy arrays
    if energy.lower() in alias.units['eV']:
        potentials = np.array(potentials) * 1000
    elif energy.lower() in alias.units['meV']:
        potentials = np.array(potentials)
    elif energy.lower() in alias.units['Ry']:
        potentials = np.array(potentials) * constants.Ry_to_meV
    else:
        raise ValueError(f"Energy unit '{energy}' not recognized.")
    # Set the system
    system.grid = np.array(positions)
    system.gridsize = len(positions)
    system.potential_values = np.array(potentials)
    # System comment as the loaded comment or the parent folder name
    if comment:
        system.comment = comment
    elif loaded_comment:
        system.comment = loaded_comment
    else:
        system.comment = os.path.basename(os.path.dirname(file_path))
    if tags:
        system.tags = tags
    print(f"Loaded {filepath}")
    return system


def from_qe(
        folder=None,
        filepath:str='potential.csv',
        include:list=['.out'],
        exclude:list=['slurm-'],
        energy:str='meV',
        comment:str=None,
        ) -> System:
    """Compiles a rotational potential CSV file from Quantum ESPRESSO pw.x outputs,
    created with `qrotor.rotation.rotate_qe()`.
    Returns a `System` object with the new potential values.

    The angle in degrees is extracted from the output filenames,
    which must follow `whatever_ANGLE.out`.

    Outputs from SCF calculations must be located in the provided `folder` (CWD if None).
    Files can be filtered by those containing the specified `include` filters,
    excluding those containing any string from the `exclude` list. 
    The output `filepath` name is `'potential.dat'` by default.

    Energy values are saved to meV by dafault, unless specified in `energy`.
    Only change the energy units if you know what you are doing;
    remember that default energy units in QRotor are meV!
    """
    folder = file.get_dir(folder)
    # Check if a previous potential.dat file exists, and ask to overwrite it
    previous_potential_file = file.get(filepath, return_anyway=True)
    if previous_potential_file:
        print(f"WARNING: Previous '{filepath}' file will be overwritten, proceed anyway?")
        answer = input("(y/n): ")
        if not answer.lower() in alias.boolean[True]:
            print("Aborted.")
            return None
    # Get the files to read
    files = file.get_list(folder=folder, include=include, exclude=exclude, abspath=True)
    folder_name = os.path.basename(folder)
    # Set header
    potential_data = f'## {comment}\n' if comment else f'## {folder_name}\n'
    potential_data += '# Rotational potential dataset\n'
    potential_data += f'# Calculated with QE pw.x using QRotor {__version__}\n'
    potential_data += '# https://pablogila.github.io/qrotor\n'
    potential_data += '#\n'
    if energy.lower() in alias.units['eV']:
        potential_data += '# Angle/deg,    Potential/eV\n'
    elif energy.lower() in alias.units['meV']:
        potential_data += '# Angle/deg,    Potential/meV\n'
    elif energy.lower() in alias.units['Ry']:
        potential_data += '# Angle/deg,    Potential/Ry\n'
    else:
        potential_data += '# Angle/deg,    Potential/meV\n'
    potential_data += '#\n'
    potential_data_list = []
    print('Extracting the potential as a function of the angle...')
    print('----------------------------------')
    counter_success = 0
    counter_errors = 0
    for file_path in files:
        filename = os.path.basename(file_path)
        file_path = file.get(filepath=file_path, include='.out', return_anyway=True)
        if not file_path:  # Not an output file, skip it
            continue
        content = pwx.read_out(file_path)
        if not content['Success']:  # Ignore unsuccessful calculations
            print(f'x   {filename}')
            counter_errors += 1
            continue
        if energy.lower() in alias.units['eV']:
            energy_value = content['Energy'] * constants.Ry_to_eV
        elif energy.lower() in alias.units['meV']:
            energy_value = content['Energy'] * constants.Ry_to_meV
        elif energy.lower() in alias.units['Ry']:
            energy_value = content['Energy']
        else:
            print(f"WARNING: Energy unit '{energy}' not recognized, using meV instead.")
            energy = 'meV'
            energy_value = content['Energy'] * constants.Ry_to_meV
        splits = filename.split('_')
        angle_value = splits[-1].replace('.out', '')
        angle_value = float(angle_value)
        potential_data_list.append((angle_value, energy_value))
        print(f'OK  {filename}')
        counter_success += 1
    # Sort by angle
    potential_data_list_sorted = sorted(potential_data_list, key=lambda x: x[0])
    # Append the sorted values as a string
    for angle_value, energy_value in potential_data_list_sorted:
        potential_data += f'{angle_value},    {energy_value}\n'
    with open(filepath, 'w') as f:
        f.write(potential_data)
    print('----------------------------------')
    print(f'Succesful calculations (OK): {counter_success}')
    print(f'Faulty calculations     (x): {counter_errors}')
    print('----------------------------------')
    print(f'Saved angles and potential values at {filepath}')
    # Warn the user if not in default units
    if energy.lower() not in alias.units['meV']:
        print(f"WARNING: You saved the potential in '{energy}' units! Remember that QRotor works in meVs!")
    new_system = None
    try:
        new_system = load(filepath=filepath, comment=comment, energy=energy)
    except:
        pass
    return new_system


def merge(
        add=[],
        subtract=[],
        comment:str=None
        ) -> System:
    """Add or subtract potentials from different systems.

    Adds the potential values from the systems in `add`,
    removes the ones from `subtract`.
    All systems will be interpolated to the bigger gridsize if needed.

    A copy of the first System will be returned with the resulting potential values,
    with an optional `comment` if indicated.
    """
    add = systems.as_list(add)
    subtract = systems.as_list(subtract)
    gridsizes = systems.get_gridsizes(add)
    gridsizes.extend(systems.get_gridsizes(subtract))
    max_gridsize = max(gridsizes)
    # All gridsizes should be max_gridsize
    for s in add:
        if s.gridsize != max_gridsize:
            s.gridsize = max_gridsize
            s = interpolate(s)
    for s in subtract:
        if s.gridsize != max_gridsize:
            s.gridsize = max_gridsize
            s = interpolate(s)

    if len(add) == 0:
        if len(subtract) == 0:
            raise ValueError('No systems were provided!')
        result = deepcopy(subtract[0])
        result.potential_values = -result.potential_values
        subtract.pop(0)
    else:
        result = deepcopy(add[0])
        add.pop(0)

    for system in add:
        result.potential_values = np.sum([result.potential_values, system.potential_values], axis=0)
    for system in subtract:
        result.potential_values = np.sum([result.potential_values, -system.potential_values], axis=0)
    if comment != None:
        result.comment = comment
    return result


def scale(
        system:System,
        factor:float,
        comment:str=None
        ) -> System:
    """Returns a copy of `system` with potential values scaled by a `factor`.

    An optional `comment` can be included.
    """
    result = deepcopy(system)
    if factor != 0:
        result.potential_values = system.potential_values * factor
    else:
        result.potential_values = np.zeros(system.gridsize)
    if comment != None:
        result.comment = comment
    return result


def interpolate(system:System) -> System:
    """Interpolates the current `System.potential_values`
    to a new grid of size `System.gridsize`.

    This basic function is called by `qrotor.solve.potential()`,
    which is the recommended way to interpolate potentials.
    """
    print(f"Interpolating potential to a grid of size {system.gridsize}...")
    V = system.potential_values
    grid = system.grid
    gridsize = system.gridsize
    new_grid = np.linspace(0, 2*np.pi, gridsize)
    # Impose periodic boundary conditions
    grid_periodic = np.append(grid, grid[0] + 2*np.pi)
    V_periodic = np.append(V, V[0])
    cubic_spline = CubicSpline(grid_periodic, V_periodic, bc_type='periodic')
    new_V = cubic_spline(new_grid)
    system.grid = new_grid
    system.potential_values = new_V
    return system


def solve(system:System):
    """Solves `System.potential_values`
    according to the `System.potential_name`,
    returning the new `potential_values`.
    Avaliable potential names are `zero`, `sine` and `titov2023`.

    If `System.potential_name` is not present or not recognised,
    the current `System.potential_values` are used.

    This basic function is called by `qrotor.solve.potential()`,
    which is the recommended way to solve potentials.
    """
    data = deepcopy(system)
    # Is there a potential_name?
    if not data.potential_name:
        if data.potential_values is None or len(data.potential_values) == 0:
            raise ValueError(f'No potential_name and no potential_values found in the system!')
    elif data.potential_name.lower() == 'titov2023':
        data.potential_values = titov2023(data)
    elif data.potential_name.lower() in alias.math['0']:
        data.potential_values = zero(data)
    elif data.potential_name.lower() in alias.math['sin']:
        data.potential_values = sine(data)
    elif data.potential_name.lower() in alias.math['cos']:
        data.potential_values = cosine(data)
    # At least there should be potential_values
    #elif not any(data.potential_values):
    elif data.potential_values is None or len(data.potential_values) == 0:
        raise ValueError("Unrecognised potential_name '{data.potential_name}' and no potential_values found")
    return data.potential_values


def zero(system:System):
    """Zero potential.

    $V(x) = 0$
    """
    x = system.grid
    return 0 * np.array(x)


def sine(system:System):
    """Sine potential.

    $V(x) = C_0 + \\frac{C_1}{2} sin(x C_2 + C_3)$  
    With $C_0$ as the potential offset,
    $C_1$ as the max potential value (without considering the offset),
    $C_2$ as the frequency, and $C_3$ as the phase.
    If no `System.potential_constants` are provided, defaults to $sin(3x)$  
    """
    x = system.grid
    C = system.potential_constants
    C0 = 0
    C1 = 1
    C2 = 3
    C3 = 0
    if C:
        if len(C) > 0:
            C0 = C[0]
        if len(C) > 1:
            C1 = C[1]
        if len(C) > 2:
            C2 = C[2]
        if len(C) > 3:
            C3 = C[3]
    return C0 + (C1 / 2) * np.sin(np.array(x) * C2 + C3)


def cosine(system:System):
    """Cosine potential.

    $V(x) = C_0 + \\frac{C_1}{2} cos(x C_2 + C_3)$  
    With $C_0$ as the potential offset,
    $C_1$ as the max potential value (without considering the offset),
    $C_2$ as the frequency, and $C_3$ as the phase.
    If no `System.potential_constants` are provided, defaults to $cos(3x)$  
    """
    x = system.grid
    C = system.potential_constants
    C0 = 0
    C1 = 1
    C2 = 3
    C3 = 0
    if C:
        if len(C) > 0:
            C0 = C[0]
        if len(C) > 1:
            C1 = C[1]
        if len(C) > 2:
            C2 = C[2]
        if len(C) > 3:
            C3 = C[3]
    return C0 + (C1 / 2) * np.cos(np.array(x) * C2 + C3)


def titov2023(system:System):
    """Potential energy function of the hindered methyl rotor, from
    [K. Titov et al., Phys. Rev. Mater. 7, 073402 (2023)](https://link.aps.org/doi/10.1103/PhysRevMaterials.7.073402).  

    $V(x) = C_0 + C_1 sin(3x) + C_2 cos(3x) + C_3 sin(6x) + C_4 cos(6x)$  
    Default constants are `qrotor.constants.constants_titov2023`[0].  
    """
    x = system.grid
    C = system.potential_constants
    if C is None:
        C = constants.constants_titov2023[0]
    return C[0] + C[1] * np.sin(3*x) + C[2] * np.cos(3*x) + C[3] * np.sin(6*x) + C[4] * np.cos(6*x)




from setuptools import setup, find_packages

DESCRIPTION = "QRotor"

exec(open('qrotor/_version.py').read())

with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name = 'qrotor', 
    version = __version__,
    author = 'Pablo Gila-Herranz',
    author_email = 'pgila001@ikasle.ehu.eus',
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    long_description_content_type = 'text/markdown',
    packages = find_packages(),
    install_requires = ['scipy', 'pandas', 'numpy', 'matplotlib', 'aton', 'periodictable'],
    extras_requires = {'dev': ['pytest', 'twine', 'build']},
    python_requires = '>=3',
    license = 'AGPL-3.0',
    keywords = ['QRotor', 'Molecular rotations', 'Quantum rotations', 'Quantum', 'Molecular', 'Rotations', 'Neutrons', 'Research', 'Ab-initio', 'DFT', 'Density Functional Theory', 'Quantum ESPRESSO', 'Phonons', 'Electronic structure'],
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Other OS",
    ]
)
